#include <stdlib.h>
#include <iostream>
#include <fstream>
#include "twsim.h"
#include "cell.h"
#include "genr.h"
#include "segment.h"
#include "transducer.h"
using namespace std;

class TrafficMessageManager: public MessageManager
{
	public:
		TrafficMessageManager(int rank, int size):
		MessageManager(),
		rank(rank),
		size(size)
		{
			int segments = size - 2;
			int genr_rank = size - 1;
			int transd_rank = size - 2;
			/// Segment to segment coupling
			for (int i = 0; i < segments-1; i++)
			{
				Address src(i,segment::out);
				Address dst(i+1,segment::in);
				couple(src,dst);
			}
			/// Segment to transducer coupling
			for (int i = 0; i < segments; i++)
			{
				Address src(i,segment::stats);
				Address dst(transd_rank,transducer::sample);
				couple(src,dst);
			}
			/// Generator to first segment coupling
			Address src(genr_rank,genr::out);
			Address dst(0,segment::in);
			couple(src,dst);
		}
		
		darray<char> pack(const PortValue& output)
		{
			// Generator
			if (rank == size-1)
				return pack_entity(*((entity*)output.value));
			// Road segment
			else if (output.port == segment::out && rank < size-3) 
				return pack_entity(*((entity*)output.value));
			else if (output.port == segment::stats && rank < size-2) 
				return pack_cell_data(*((cell_data*)output.value));
			else
				return darray<char>(0);
		}

		PortValue unpack(const char* buf, const Address& src, const Address& dst)
		{
			// Segment
			if (dst.rank < size - 2)
			{
				entity* e = new entity(unpack_entity(buf));
				return PortValue(dst.port,e);
			}
			// Transducer
			else if (dst.rank == size - 2)
			{
				cell_data* d = new cell_data(unpack_cell_data(buf));
				return PortValue(dst.port,d);
			}
			else
				return PortValue();
		}

		~TrafficMessageManager()
		{
		}

		void message_gc(PortValue& pv)
		{
			delete pv.value;
		}

	private:

		int rank;
		int size;

		darray<char> pack_entity(const entity& e) 
		{
			double eSize = e.get_size();
			return darray<char>((const char*)&eSize, sizeof(double));
		}
		darray<char> pack_cell_data(const cell_data& d) 
		{
			darray<char> buffer;
			int ID = d.getID();
			int size = d.getSize();
			buffer.append((const char*)&ID,sizeof(int));
			buffer.append((const char*)&size,sizeof(int));
			return buffer;
		}
		entity unpack_entity(const char* x) 
		{
			double eSize;
			memcpy((char*)&eSize,x,sizeof(double));
			return entity(eSize);
		}
		cell_data unpack_cell_data(const char* x) 
		{
			int size, ID;
			memcpy((char*)&ID,x,sizeof(int));
			memcpy((char*)&size,x+sizeof(int),sizeof(int));
			return cell_data(ID,size);
		}
};

int main(int argc, char** argv) 
{
	ami::init(argc,argv);
	devs* d = NULL;
	long seed, cells;
	int procs;
	double speed_mean, entry_mean;
	double sample_time, running_time;
	ifstream params ("input");
	params >> speed_mean;
	params >> entry_mean;
	params >> cells;
	params >> sample_time;
	params >> running_time;
	params >> seed;
	params >> procs;
	params.close();

	int segments = procs - 2;
	int cells_per_proc = div(cells,segments).quot;
	int left_over = div(cells,segments).rem;
	if (ami::get_rank() == segments - 1) 
	{
		segment* model = new segment(cells_per_proc+left_over,
		ami::get_rank()*cells_per_proc);
		d = model;
	}
	else if (ami::get_rank() < segments) 
	{
		segment* model = new segment(cells_per_proc,
		ami::get_rank()*cells_per_proc);
		d = model;
	}
	else if (ami::get_rank() == procs - 2) 
	{
		transducer* model = new transducer(sample_time);
		d = model;
	}
	else if (ami::get_rank() == procs - 1)
	{
		genr* model = new genr(entry_mean,speed_mean,seed);
		d = model;
	}
	twsim sim(d,new TrafficMessageManager(ami::get_rank(),ami::get_size()));
	sim.run(running_time);
	transducer* model = dynamic_cast<transducer*>(sim.getModel());
	if (model != NULL) model->print_report();
	ami::fini();
	return 0;
}
