/***************
adevs, a library supporting discrete event simulation
Copyright (C) 2001 James Nutaro

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#ifndef __qode_h_
#define __qode_h_
#include <qintg.h>
/**
A quantized ODE solver.  The qode uses a set of quantized integrators
to provide a solution to a set of ordinary differential equations.
*/
class qode 
{
	public:
		/// Create a solver for a set of N ODEs.
		qode(unsigned N);
		/// Copy constructor
		qode(const qode& src);
		/// Assigment operator
		const qode& operator=(const qode& src);
		/**
		Initial the integrator set.  q0 is the initial state, dq0 are
		the initial derivatives.
		*/
		void init(const double* q0, const double* dq0);
		/// Initialize the jth integrator
		void init(unsigned j, double q0, double dq0 = 0.0) { q[j].init(q0,dq0); }
		/**
		Integrate over dt (dt <= ta()) units of time with derivatives dq.  
		*/
		void integ(double dt, const double* dq);
		/**
		Integrate over dt (dt <= ta()) units of time with unchanged derivatives.  
		*/
		void integ(double dt);
		/**
		Return the quantized output for integrator j at time tL + ta(),
		where tL was the time of the last state change.
		*/
		double out(unsigned j) const { return q[j].out(sigma); }
		/// Return the quantized output for integrator j at time tL + dt, dt <= ta()
		double out(unsigned j, double dt) const { return q[j].out(dt); }
		/// Return the quantized state of the integrator at time tL
		double qstate(unsigned j) const { return q[j].qstate(); }
		/// Return the state of the integrator at time tL
		double state(unsigned j) const { return q[j].state(); }
		/// The time advance function for the ODE set. 
		double ta() const { return sigma; }
		/// Set the parameters for iterator j (see qintg)
		void setParams(unsigned j, double D, double e = -1.0) 
		{
			q[j].setParams(D,e);
		}
		/// Set the parameters for all integrators (see qintg)
		void setParams(double D, double e = -1.0);
	private:
		// Integrator set
		qintg* q;
		// Time to next boundary crossing
		double sigma;
		// Number of integrators
		unsigned N;

		void compute_ta();
};

#endif
