/***************
adevs, a library supporting discrete event simulation
Copyright (C) 2002 James Nutaro

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#include <pdevs.h>

TIME_TYPE basic_pdevs_sim::granule = EPSILON;

void pdevs_coord::detachModel()
{
	// clean up the I/O bags and receiver bags.  Garbage collection
	// on these bags will be done when the atomic models that generated
	// the output are detached.
	x.removeAll();
	y.removeAll();
	receivers.removeAll();
	// Detach all component models 
	for (int i = 0; i < components.getSize(); i++)
	{
		if (components[i]->getAttachedSimulator() != NULL)
			detach_sim((basic_pdevs_sim*)(components[i]->getAttachedSimulator()));
	}
	// Finish cleaning up
	components.removeAll();
	sched.clear();
	imm.removeAll();
	model->setAttachedSimulator(NULL);
	model = NULL;
	setModel(model);
}

void pdevs_coord::attachModel(devsn* d)
{
	// Make sure that there isn't already a model attached to this simulator
	if (model != NULL)
		THROW_BASE_EXCEPTION(
		"Did not detach model from coordinator before attaching new model");
	// Attach the model
	model = d;
	setModel(model);
	model->setAttachedSimulator(this);
	// Schedule all of the component models.  The time of last event is the
	// smallest of the component last events.  The time of next event is
	// the smallest time of next event of all the component models.
	// Don't need to worry about current time since that will be set in 
	// the computeDeltaFunc(t) call and isn't needed before then.
	delta_struct = NULL;
	if (model->getNetExec()->timeNext() < INFINITY)
			sched.insert(attach_sim(model->getNetExec()),model->getNetExec()->timeNext());
	model->setTimeLast(model->getNetExec()->timeLast());
	model->getNetExec()->getComponents(components);
	for (int i = 0; i < components.getSize(); i++)
	{
		if (components[i]->timeNext() < INFINITY)
			sched.insert(attach_sim(components[i]),components[i]->timeNext());
		if (components[i]->timeLast() < model->timeLast())
			model->setTimeLast(components[i]->timeLast());
	}
	if (sched.getSize() > 0)
		model->setTimeNext(sched.minPriority());
	else
		model->setTimeNext(INFINITY);
}

void pdevs_coord::computeInputOutput()
{
	// No imminent models, so no output
	if (sched.getSize() == 0) return;
	// Get all of the imminent simulators
	TIME_TYPE tN = sched.minPriority();
	while (sched.minPriority() - tN <= granule)
	{
		imm.add(sched.getMinimum());
		sched.removeMinimum();
		if (sched.getSize() == 0)
			break;
	}
	// Route the output from the imminent simulators
	int num_imm = imm.getSize();
	for (int i = 0; i < num_imm; i++)
	{
		basic_pdevs_sim* sim = imm[i];
		sim->computeInputOutput();
		for (int k = 0; k < sim->getOutput().getSize(); k++)
		{
			receivers.removeAll();
			PortValue pv(sim->getOutput().get(k));
			model->getNetExec()->route(pv,sim->getModel(),receivers);
			for (int j = 0; j < receivers.getSize(); j++)
			{
				// external output
				if (receivers[j].model == model) 
					y.add(PortValue(receivers[j].port,pv.value));
				// internal model I/O (no direct feedback allowed)
				else if (receivers[j].model != sim->getModel())
				{
					basic_pdevs_sim* rsim = attach_sim(receivers[j].model);
					rsim->inject(PortValue(receivers[j].port,pv.value));
					imm.add(rsim);
				}
				// any other coupling is in error
				else
					THROW_BASE_EXCEPTION("Model tried to influence self");
			}
		}
	}
}				

void pdevs_coord::computeDeltaFunc(TIME_TYPE t)
{
	// Check the request time of next event for validity
	if (t < timeLast() || t > timeNext()+granule)
		THROW_BASE_EXCEPTION("Invalid time of next event");
	// If t is too early, we have no imminent models
	// (remember, computeInputOutput assumes t = tN)
	// so place all models imminent at tN back into the
	// schedule
	if (timeNext() - t > granule)
	{
		for (int i = 0; i < imm.getSize(); i++)
		{
			devs* d = imm[i]->getModel();
			// detach the model to reset the simulator state
			detach_sim(imm[i]);
			// re-attach the model if its not passive
			if (d->timeNext() < INFINITY)
				sched.insert(attach_sim(d),d->timeNext());
		}
		imm.removeAll();
	}

	// Route external input
	for (int i = 0; i < x.getSize(); i++)
	{
		receivers.removeAll();
		model->getNetExec()->route(x[i],model,receivers);
		for (int j = 0; j < receivers.getSize(); j++)
		{
			if (receivers[j].model == model)
				THROW_BASE_EXCEPTION("Direct input to output coupling in network model")
			// external input
			else
			{
				basic_pdevs_sim* rsim = attach_sim(receivers[j].model);
				rsim->inject(PortValue(receivers[j].port,x[i].value));
				imm.add(rsim);
			}
		}
	}
	// Compute next state for all imminent models.  Network executive
	// always goes last.
	model->setTimeCurrent(t);
	for (int i = 0; i < imm.getSize(); i++)
	{
		if (imm[i]->getModel() != model->getNetExec())
			imm[i]->computeDeltaFunc(t);
		else
			delta_struct = imm[i];
	}
	// If network exec is imminent, compute its next state.
	// New model structure is accounted for during the endOfCycle() call.
	if (delta_struct != NULL)
		delta_struct->computeDeltaFunc(t);
	// Clear the external input list (they have been consumed)
	x.removeAll();
	model->setTimeLast(t);
}

void pdevs_coord::endOfCycle()
{
	// Schedule the models that changed state
	for (int i = 0; i < imm.getSize(); i++)
	{
		// Garbage collection and scheduling of component coupled models takes
		// place during endOfCycle() call
		imm[i]->endOfCycle();
		// Schedule the component (handle network model later)
		if (imm[i]->timeNext() < INFINITY)
			sched.insert(imm[i],imm[i]->timeNext());
		else
			detach_sim(imm[i]);
	}
	// Account for structure change and reschedule the network executive
	if (delta_struct != NULL)
	{
		// Get the new component set
		adevs_set<devs*> new_components(components.getCapacity());
		model->getNetExec()->getComponents(new_components);
		// Find the removed components by subtracting the new components
		// from the old component set.  What's left are the components
		// that are not in the new component set.
		adevs_set<devs*> removed(components - new_components);
		// Detach simulators attached to the removed models
		for (int i = 0; i < removed.getSize(); i++)
		{
			if (removed[i]->getAttachedSimulator() != NULL)
				detach_sim((basic_pdevs_sim*)removed[i]->getAttachedSimulator());
		}
		// Tell the network executive to clean up the removed models
		model->getNetExec()->gc_models(removed);
		// Find the new components
		adevs_set<devs*> added(new_components - components);
		// Initialize and schedule all of the new components
		for (int i = 0; i < added.getSize(); i++)
		{
			added[i]->setTimeCurrent(model->timeCurrent());
			added[i]->setTimeLast(model->timeCurrent());
			added[i]->init();
			if (added[i]->timeNext() < INFINITY)
				sched.insert(attach_sim(added[i]),added[i]->timeNext());
		}
		// Update the persistent component set
		components.removeAll();
		components.append(new_components);
		// Reschedule the network executive
		delta_struct = NULL;
	}
	// Set the network executives time of next event
	if (sched.getSize() > 0)
		model->setTimeNext(sched.minPriority());
	else
		model->passivate();
	// Done.  Clean out the output set (which is no longer valid for the model
	// state) and the imminent set.
	imm.removeAll();
	y.removeAll();
}

void pdevs_coord::detach_sim(basic_pdevs_sim* sim)
{
	// remove the simulator from the schedule
	if (sched.isEnqueued(sim))
		sched.remove(sim);
	// detach the model
	sim->detachModel();
	// return the simulator to the appropriate pool
	if (sim->getType() == SIM)
		pdevs_sims.add((pdevs_sim*)sim);
	else
		pdevs_coords.add((pdevs_coord*)sim);
}

basic_pdevs_sim* pdevs_coord::attach_sim(devs* d)
{
	basic_pdevs_sim* sim = NULL;
	// If a simulator is already attached to the model, get that simulator
	if (d->getAttachedSimulator() != NULL)
		sim = (basic_pdevs_sim*)(d->getAttachedSimulator());
	// otherwise, if a simulator already exists in the appropriate
	// pool of simulators, get it out of there and attach the model to it
	else if (d->getType() == devs::ATOMIC && pdevs_sims.getSize() > 0)
	{
		pdevs_sim* asim = pdevs_sims[0];
		pdevs_sims.remove(0);
		asim->attachModel((atomic*)d);
		sim = asim;
	}
	// As a last resort, create a new simulator object for this model
	else if (d->getType() == devs::ATOMIC && pdevs_sims.getSize() == 0)
	{
		pdevs_sim* asim = new pdevs_sim;
		asim->attachModel((atomic*)d);
		sim = asim;
	}
	else if (d->getType() == devs::NETWORK && pdevs_coords.getSize() > 0)
	{
		pdevs_coord* csim = pdevs_coords[0];
		pdevs_coords.remove(0);
		csim->attachModel((devsn*)d);
		sim = csim;
	}
	else
	{
		pdevs_coord* csim = new pdevs_coord;
		csim->attachModel((devsn*)d);
		sim = csim;
	}
	return sim;
}

pdevs_coord::~pdevs_coord()
{
	for (int i = 0; i < imm.getSize(); i++)
		detach_sim(imm[i]);
	for (int i = 0; i < pdevs_sims.getSize(); i++)
		delete pdevs_sims[i];
	for (int i = 0; i < pdevs_coords.getSize(); i++)
		delete pdevs_coords[i];
}
